#include "game/phase/phase_demo.h"

#include "game/demo.h"
#include "game/fader.h"
#include "game/game.h"
#include "game/interpolation.h"
#include "game/inventory_ring.h"
#include "game/output.h"
#include "game/shell.h"
#include "game/text.h"
#include "memory.h"

typedef enum {
    STATE_RUN,
    STATE_FADE_OUT,
    STATE_FINISH,
} STATE;

typedef struct {
    STATE state;
    int32_t level_num;
    FADER top_fader;
    GF_COMMAND exit_gf_cmd;
} M_PRIV;

static PHASE_CONTROL M_Start(PHASE *phase);
static void M_End(PHASE *phase);
static void M_Suspend(PHASE *const phase);
static void M_Resume(PHASE *const phase);
static PHASE_CONTROL M_Control(PHASE *phase, int32_t n_frames);
static void M_Draw(PHASE *phase);

static PHASE_CONTROL M_Start(PHASE *const phase)
{
    M_PRIV *const p = phase->priv;
    if (p->level_num == -1) {
        return (PHASE_CONTROL) {
            .action = PHASE_ACTION_END,
            .gf_cmd = { .action = GF_EXIT_TO_TITLE },
        };
    }

    if (!Demo_Start(p->level_num)) {
        return (PHASE_CONTROL) {
            .action = PHASE_ACTION_END,
            .gf_cmd = { .action = GF_EXIT_TO_TITLE },
        };
    }

    p->state = STATE_RUN;
    Game_SetIsPlaying(true);

    return (PHASE_CONTROL) { .action = PHASE_ACTION_CONTINUE };
}

static void M_End(PHASE *const phase)
{
    Demo_End();
}

static void M_Suspend(PHASE *const phase)
{
    Game_SetIsPlaying(false);
    Demo_Pause();
}

static void M_Resume(PHASE *const phase)
{
    Game_SetIsPlaying(true);
    Demo_Unpause();
}

static PHASE_CONTROL M_Control(PHASE *const phase, const int32_t num_frames)
{
    M_PRIV *const p = phase->priv;

    switch (p->state) {
    case STATE_RUN:
        for (int32_t i = 0; i < num_frames; i++) {
            const GF_COMMAND gf_cmd = Demo_Control();
            if (gf_cmd.action != GF_NOOP) {
                p->state = STATE_FADE_OUT;
                p->exit_gf_cmd = gf_cmd;
                Fader_Init(&p->top_fader, FADER_ANY, FADER_BLACK, 0.5);
                return (PHASE_CONTROL) { .action = PHASE_ACTION_NO_WAIT };
            }
        }

        break;

    case STATE_FADE_OUT:
        Game_SetIsPlaying(false);
        Demo_StopFlashing();
        if (!Fader_IsActive(&p->top_fader)) {
            p->state = STATE_FINISH;
            return (PHASE_CONTROL) { .action = PHASE_ACTION_NO_WAIT };
        }
        break;

    case STATE_FINISH:
        return (PHASE_CONTROL) {
            .action = PHASE_ACTION_END,
            .gf_cmd = Shell_IsExiting()
                ? (GF_COMMAND) { .action = GF_EXIT_GAME }
                : p->exit_gf_cmd,
        };
    }

    return (PHASE_CONTROL) { .action = PHASE_ACTION_CONTINUE };
}

static void M_Draw(PHASE *const phase)
{
    M_PRIV *const p = phase->priv;
    if (p->state == STATE_FADE_OUT) {
        Interpolation_Disable();
    }
    Game_Draw(true);
    if (p->state == STATE_FADE_OUT) {
        Interpolation_Enable();
    }
    Text_Draw();
    Fader_Draw(&p->top_fader);
    Output_DrawPolyList();
}

PHASE *Phase_Demo_Create(const int32_t level_num)
{
    PHASE *const phase = Memory_Alloc(sizeof(PHASE));
    M_PRIV *const p = Memory_Alloc(sizeof(M_PRIV));
    p->level_num = level_num;
    phase->priv = p;
    phase->start = M_Start;
    phase->end = M_End;
    phase->suspend = M_Suspend;
    phase->resume = M_Resume;
    phase->control = M_Control;
    phase->draw = M_Draw;
    return phase;
}

void Phase_Demo_Destroy(PHASE *const phase)
{
    M_PRIV *const p = phase->priv;
    Memory_Free(p);
    Memory_Free(phase);
}
